# -*- coding: utf-8 -*-
import logging
import os
import sys

import datetime

import six

from shire.const import SHIRE_WORKHORSE_PROCESS_NAME


__all__ = [
    'to_list', 'check_pid_is_shire', 'capture_output', 'cleanup_old_jobs', 'execute_job', 'create_console_handler',
    'activate_venv', 'is_venv',
]


def create_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    return logger


def create_console_handler(logger, fmt='[ %(asctime)s | %(name)s ] - <%(levelname)s> - %(message)s'):
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)
    console.setFormatter(logging.Formatter(fmt))
    logger.addHandler(console)


def to_list(value, default=None):
    if value is None:
        return default if default else []
    if isinstance(value, six.string_types):
        return value.split(',')
    if isinstance(value, (list, tuple)):
        return map(str, value)
    return [str(value), ]


def check_pid_is_shire(pid):
    # Проверяет, что процесс жив и процесс - это shire workhorse
    proc = '/proc/{}/cmdline'.format(pid)
    if os.path.exists(proc):
        with open(proc, 'rb') as f:
            return f.read().startswith(SHIRE_WORKHORSE_PROCESS_NAME)
    return False


class capture_output(object):
    def __init__(self, stdout=None, stderr=None):
        self._stdout = stdout or sys.stdout
        self._stderr = stderr or sys.stderr

    def __enter__(self):
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.old_stdout.flush()
        self.old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr

    def __exit__(self, exc_type, exc_value, traceback):
        self._stdout.flush()
        self._stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr


def cleanup_old_jobs(days_ago=7):
    from shire.models import JobEntry
    days_ago = datetime.date.today() - datetime.timedelta(days=days_ago)
    query = JobEntry.delete().where(~(JobEntry.updated_at >> None) & (JobEntry.updated_at < days_ago))
    return query.execute()


def execute_job(job_id, config):
    from shire.models import JobEntry
    from shire.job import Job
    job_entry = JobEntry.select().where(JobEntry.id == job_id).first()
    if job_entry is None:
        return False
    args, kwargs = job_entry.get_params()
    # Подготавливаем job_entry к inline-запуску
    job_entry.pool = 'dummy_pool'
    job_entry.queue = 'dummy_queue'
    job_entry.host = 'dummy_host'
    job_entry.status = JobEntry.STATUS_ENQUEUED
    job_entry.save()
    result = Job.execute(config, args=args, kwargs=kwargs, job_entry=job_entry)
    return result.status == JobEntry.STATUS_ENDED


def decode_if_not_empty(value):
    if isinstance(value, six.binary_type):
        return value.decode()
    return value


def activate_venv(path, venv_exclusive=None):
    if any(x.startswith(path) for x in sys.path):
        return

    # как активировать окружение изнутри? даже если другой пайтон? и нужно ли...
    # from py._builtin import execfile
    # activate_script = os.path.join(path, "bin", "activate_this.py")
    # execfile(activate_script, dict(__file__=activate_script))
    # exec(compile(open(filename, "rb").read(), filename, 'exec'), globals, locals)

    old_os_path = os.environ.get('PATH', '')
    os.environ['PATH'] = os.path.dirname(os.path.join(path, 'bin')) + os.pathsep + old_os_path
    site_packages = os.path.join(path, 'lib', 'python%s' % sys.version[:3], 'site-packages')
    if venv_exclusive:
        for item in list(sys.path):
            if 'site-packages' in item:
                sys.path.remove(item)
    prev_sys_path = list(sys.path)
    import site
    site.addsitedir(site_packages)
    sys.real_prefix = sys.prefix
    sys.prefix = path
    # Move the added items to the front of the path:
    new_sys_path = []
    for item in list(sys.path):
        if item not in prev_sys_path:
            new_sys_path.append(item)
            sys.path.remove(item)
    sys.path[:0] = new_sys_path


def is_venv():
    return (hasattr(sys, 'real_prefix') or
            (hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix))



# -*- coding: utf-8 -*-
import json
import logging
from logging.handlers import RotatingFileHandler
import os
import sys

import datetime
from shire.redis_managers import LogMessageManager
from shire.utils import create_console_handler, create_logger


__all__ = ['Scribe']


class Scribe(object):
    POP_TIMEOUT = 1

    # Режимы работы Scribe в зависимости от конфигурации
    MODE_ONE_LOG, MODE_MULTI_LOGS = 0, 1

    SYSTEM_LOG = 'system'
    DEFAULT_LOG = 'default'

    FORMATTER = logging.Formatter('%(message)s')
    OUR_FORMATTER = u'[ %(asctime)s | %(name)s | %(pathname)s:%(lineno)d ] - <%(levelname)s> - %(message)s'

    EXPIRED_NOTIFY_LIMIT = 500

    def __init__(self, config, verbose=False):
        self.config = config
        self.verbose = verbose
        self.scribe_logger = create_logger('shire.scribe_proccess')
        if verbose:
            create_console_handler(self.scribe_logger)
        self.section = self.config.section_getter(self.config.SCRIBE_SECTION)
        self.manager = LogMessageManager(self.config.get_redis())
        self.log_dir = self.section.get(self.config.SCRIBE_DIRECTORY, self.config.SCRIBE_DIRECTORY_DEFAULT)
        self.max_logs = int(self.section.get(self.config.SCRIBE_MAX_LOGS, self.config.SCRIBE_MAX_LOGS_DEFAULT))
        self.log_size = int(self.section.get(self.config.SCRIBE_MAX_SIZE, self.config.SCRIBE_MAX_SIZE_DEFAULT))
        self.handlers = {}
        self.loggers = {}
        self._expired_logs_count = 0
        self.setup_handlers()
        self.log_start()

    def scribe_log(self, msg, level='info'):
        if not self.verbose:
            return
        getattr(self.scribe_logger, level)(msg)

    @property
    def expired_logs_count(self):
        return self._expired_logs_count

    @expired_logs_count.setter
    def expired_logs_count(self, value):
        self._expired_logs_count = value
        if value % self.EXPIRED_NOTIFY_LIMIT == 1:
            self.notify_expire_count()

    def get_full_path(self, filename):
        return os.path.join(self.log_dir, filename)

    @property
    def mode(self):
        if self.section.get(self.config.SCRIBE_PER_POOL, self.config.SCRIBE_PER_POOL_DEFAULT) != '1':
            return self.MODE_ONE_LOG
        return self.MODE_MULTI_LOGS

    @property
    def system(self):
        return self.loggers[self.SYSTEM_LOG]

    def make_logger(self, name):
        logger = logging.getLogger('shire.scribe.{}'.format(name))
        logger.setLevel(logging.DEBUG)
        self.loggers[name] = logger
        return logger

    def make_handler(self, name, path):
        handler = RotatingFileHandler(path, maxBytes=self.log_size, backupCount=self.max_logs)
        handler.setFormatter(self.FORMATTER)
        handler.setLevel(logging.DEBUG)
        self.handlers[name] = handler
        return handler

    def make_logger_and_handler(self, name, path):
        logger = self.make_logger(name)
        handler = self.make_handler(name, path)
        logger.addHandler(handler)

    def get_pool_logger(self, name):
        if self.mode == self.MODE_ONE_LOG:
            return self.loggers[self.DEFAULT_LOG]
        logger_name = 'pool.{}'.format(name)
        if logger_name not in self.loggers:
            self.make_logger_and_handler(logger_name, self.get_full_path('{}.log'.format(name)))
        return self.loggers[logger_name]

    def setup_handlers(self):
        self.make_logger_and_handler(self.SYSTEM_LOG, self.get_full_path(
            self.section.get(self.config.SCRIBE_SYSTEM_LOG, self.config.SCRIBE_SYSTEM_LOG_DEFAULT)
        ))
        if self.mode == self.MODE_ONE_LOG:
            self.make_logger_and_handler(self.DEFAULT_LOG, self.get_full_path(
                self.section.get(self.config.SCRIBE_DEFAULT_LOG, self.config.SCRIBE_DEFAULT_LOG_DEFAULT)
            ))

    def log_start(self):
        self.system.info('Scribe process started')

    def notify_expire_count(self):
        self.loggers[self.SYSTEM_LOG].info('Expired logs count: {}'.format(self.expired_logs_count))

    def write_log(self, log):
        log_data = self.manager.pop(log['pool'], log['uuid'])
        if not log_data:
            self.expired_logs_count += 1
            self.scribe_log('Got expired log for pool {}'.format(log['pool']))
            return
        log_data = json.loads(log_data)
        log_data['asctime'] = datetime.datetime.fromtimestamp(log_data['created'])
        # Форматируем оригинальное сообщение лога
        message = self.OUR_FORMATTER % log_data
        if 'exc_text' in log_data:
            message = u'{}\n{}'.format(message, log_data['exc_text'])
        logger = self.get_pool_logger(name=log['pool'])
        level = logging._checkLevel(log_data['levelname'])
        logger.log(level, message)

    def run(self):
        self.scribe_log('Scribe started')
        self.scribe_log(
            'Scribe mode: {}'.format(
                {self.MODE_MULTI_LOGS: 'multi logs', self.MODE_ONE_LOG: 'single log'}.get(self.mode)
            )
        )
        self.scribe_log('Logs directory: {}'.format(self.log_dir))
        try:
            while True:
                log = self.manager.pop_next(self.POP_TIMEOUT)
                if log:
                    self.write_log(log)
        except Exception as e:
            self.system.exception(e)
            self.system.error('Scribe process terminated')
            self.scribe_log(e, 'exception')
            sys.exit(1)



# -*- coding: utf-8 -*-
import datetime
import json
import sys
import os
import peewee

if sys.version_info < (3, 5):
    # Python 2 и старые версии 3
    import imp

    def import_from_source(fake_name, path):
        return imp.load_source(fake_name, path)
else:
    import importlib.util

    def import_from_source(fake_name, path):
        spec = importlib.util.spec_from_file_location(fake_name, path)
        _module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(_module)
        sys.modules[fake_name] = _module
        return _module


__all__ = ['db', 'Limit', 'JobEntry', 'Queue']


db = peewee.Proxy()


class BaseModel(peewee.Model):
    class Meta:
        database = db


class Limit(BaseModel):
    ENTITY_POOL = 'pool'
    ENTITY_QUEUE = 'queue'
    ENTITY_CHOICES = ((ENTITY_POOL, u'Пул'), (ENTITY_QUEUE, u'Очередь'))

    entity_type = peewee.CharField(max_length=16, choices=ENTITY_CHOICES)
    entity = peewee.CharField(max_length=256)
    limit = peewee.IntegerField(default=1)

    class Meta:
        db_table = 'shire_limit'


class Queue(BaseModel):
    # TODO проверить использование - очереди у нас виртуальные и живут в неймспейсе пула, лимиты на них в Limit
    name = peewee.CharField(max_length=256)
    pool = peewee.CharField(max_length=128)

    class Meta:
        db_table = 'shire_queue'


class JobEntry(BaseModel):
    STATUS_NEW = 'new'
    STATUS_RESTART = 'restart'
    STATUS_ENQUEUED = 'enqueued'
    STATUS_IN_PROGRESS = 'in_progress'
    STATUS_ENDED = 'ended'
    STATUS_CHOICES = (
        (STATUS_NEW, u'новая'),
        (STATUS_RESTART, u'на перезапуск'),
        (STATUS_ENQUEUED, u'в очереди'),
        (STATUS_IN_PROGRESS, u'выполняется'),
        (STATUS_ENDED, u'завершена'),
    )
    STATUS_DEFAULT = STATUS_NEW
    VALID_STATUSES = {k for k, v in STATUS_CHOICES}
    FUNC_CALL_PATH = 'path'
    FUNC_CALL_CLASS = 'class'
    FUNC_CALL_ARGS = 'args'
    FUNC_CALL_KWARGS = 'kwargs'
    FUNC_CALL_SYS_PATH = 'sys_path'
    FUNC_CALL_VENV_PATH = 'venv_path'
    FUNC_CALL_VENV_EXCLUSIVE = 'venv_exclusive'
    HOST_DEFAULT = 'default'

    func_call_ = peewee.TextField(db_column='func_call')
    pool = peewee.CharField(max_length=256, index=True)
    queue = peewee.CharField(max_length=256)
    status = peewee.CharField(max_length=32, choices=STATUS_CHOICES, default=STATUS_DEFAULT, index=True)
    host = peewee.CharField(max_length=256, default=HOST_DEFAULT, index=True)
    pool_uuid = peewee.CharField(max_length=64, null=True, index=True)
    worker_uuid = peewee.CharField(max_length=64, null=True)
    worker_pid = peewee.IntegerField(default=0, null=True)
    created_at = peewee.DateTimeField(default=datetime.datetime.now)
    execute_at = peewee.DateTimeField(default=datetime.datetime.now, index=True)
    updated_at = peewee.DateTimeField(default=None, null=True, index=True)
    ended_at = peewee.DateTimeField(default=None, null=True)

    class Meta:
        db_table = 'shire_job'

    @property
    def func_call(self):
        return json.loads(self.func_call_)

    @func_call.setter
    def func_call(self, value):
        self.func_call_ = json.dumps(value)

    @classmethod
    def create_job(cls, file_path, file_cls, pool, queue=None, status=None, host=None, args=None, kwargs=None,
                   sys_path=None, venv_path=None, venv_exclusive=None, wait_minutes=0):
        status = status if status in cls.VALID_STATUSES else cls.STATUS_DEFAULT
        host = cls.HOST_DEFAULT if host is None else host
        job = cls(pool=pool, queue=queue, status=status, host=host)
        job.func_call = cls.make_func_call(
            file_path=file_path, file_cls=file_cls, args=args, kwargs=kwargs,
            sys_path=sys_path, venv_path=venv_path, venv_exclusive=venv_exclusive)
        if wait_minutes > 0:
            job.execute_at = datetime.datetime.now() + datetime.timedelta(minutes=wait_minutes)
        return job

    @classmethod
    def make_func_call(cls, file_path, file_cls, args=None, kwargs=None,
                       sys_path=None, venv_path=None, venv_exclusive=None):
        func_call = {
            cls.FUNC_CALL_PATH: file_path,
            cls.FUNC_CALL_CLASS: file_cls,
            cls.FUNC_CALL_ARGS: () if args is None else args,
            cls.FUNC_CALL_KWARGS: {} if kwargs is None else kwargs
        }
        if sys_path is not None:
            func_call[cls.FUNC_CALL_SYS_PATH] = sys_path
        if venv_path is not None:
            func_call[cls.FUNC_CALL_VENV_PATH] = venv_path
            if venv_exclusive is not None:
                func_call[cls.FUNC_CALL_VENV_EXCLUSIVE] = venv_exclusive
        return func_call

    @property
    def job_cls(self):
        module_path = self.func_call[self.FUNC_CALL_PATH]
        if module_path.endswith('.pyc'):
            module_path = module_path[:-1]
        module_name = 'shire_fake_module{}'.format(module_path.rsplit('.', 1)[0].replace(os.path.sep, '_'))
        fake_module = import_from_source(module_name, module_path)
        return getattr(fake_module, self.func_call[self.FUNC_CALL_CLASS])

    def get_params(self):
        # параметры для конкретной задачи
        args, kwargs = [], {}
        if self.FUNC_CALL_ARGS in self.func_call:
            args = self.func_call[self.FUNC_CALL_ARGS]
            if not isinstance(args, list):
                raise Exception('func_call args {} must be list!'.format(args))
        if self.FUNC_CALL_KWARGS in self.func_call:
            kwargs = self.func_call[self.FUNC_CALL_KWARGS]
            if not isinstance(kwargs, dict):
                raise Exception('func_call kwargs {} must be dict!'.format(kwargs))
        return args, kwargs

    def set_status(self, status):
        self.status = status
        self.save(only=[JobEntry.status])

    def save(self, force_insert=False, only=None):
        if 'updated_at' not in self._dirty:
            if only is not None and self.__class__.updated_at not in only:
                only = [x for x in only] + [self.__class__.updated_at]
            self.updated_at = datetime.datetime.now()
        return super(self.__class__, self).save(force_insert=force_insert, only=only)


class Crontab(BaseModel):
    key = peewee.CharField(max_length=250, index=True)
    cron_string = peewee.TextField()
    description = peewee.TextField(null=True)
    last_scheduled = peewee.DateTimeField(null=True)

    func_call_ = peewee.TextField(db_column='func_call')
    pool = peewee.CharField(max_length=256, index=True)
    queue = peewee.CharField(max_length=256)
    host = peewee.CharField(max_length=256, default=JobEntry.HOST_DEFAULT, index=True)

    class Meta:
        db_table = 'shire_crontab'

    @property
    def func_call(self):
        return json.loads(self.func_call_)

    @func_call.setter
    def func_call(self, value):
        self.func_call_ = json.dumps(value)



# -*- coding: utf-8 -*-

import uuid

from shire.exceptions import PoolInvalidStatusException

__all__ = ['QueueManager', 'PoolStatusManager', 'LogMessageManager']


class BaseRedisManager(object):

    def __init__(self, connection):
        self.db = connection


class QueueManager(BaseRedisManager):
    PATH = 'shire:to_execute:{pool}'

    def push(self, pool, job_id, to_tail=False):
        key = self.PATH.format(pool=pool)
        if to_tail:
            return self.db.rpush(key, str(job_id))
        return self.db.lpush(key, str(job_id))

    def pop(self, pool, timeout=None):
        key = self.PATH.format(pool=pool)
        res = self.db.brpop(key, timeout=timeout)
        if res:
            key, job_id = res
            return job_id.decode()
        return None

    def show_queue(self, pool):
        return self.db.lrange(self.PATH.format(pool=pool), 0, -1)


class PoolStatusManager(BaseRedisManager):
    BASE_PATH = 'shire:pool'
    POOL_PATH = BASE_PATH + ':{pool}'
    UUID_PATH = POOL_PATH + ':{uuid}'

    STATUS_ACTIVE = 'active'
    STATUS_DEAD = 'dead'
    STATUS_KILL = 'kill'
    STATUS_TERMITATED = 'terminated'
    _valid_statuses = (STATUS_ACTIVE, STATUS_DEAD, STATUS_KILL, STATUS_TERMITATED)

    def get_all(self, pool=None):
        keys = self.POOL_PATH.format(pool=pool) + ':*' if pool else self.BASE_PATH + ':*'
        value = self.db.keys(keys)
        if value:
            return [v.decode() for v in value]
        return value

    def get_status(self, pool, _uuid):
        value = self.db.hget(self.UUID_PATH.format(pool=pool, uuid=_uuid), 'status')
        if value:
            return value.decode()
        return value

    def set_status(self, pool, _uuid, status):
        if status not in self._valid_statuses:
            raise PoolInvalidStatusException(status)
        return self.db.hset(self.UUID_PATH.format(pool=pool, uuid=_uuid), 'status', status)

    def del_status(self, pool, _uuid):
        return self.db.delete(self.UUID_PATH.format(pool=pool, uuid=_uuid))


class LogMessageManager(BaseRedisManager):
    BASE_PATH = 'shire:log:{pool}'
    UUID_PATH = BASE_PATH + ':{uuid}'
    QUEUE_PATH = 'shire:lqueue'
    ID_FORMAT = '{pool}:{uuid}'
    EXPIRE_TIME = 5 * 60

    def write(self, pool, message):
        _uuid = uuid.uuid4()
        self.db.set(self.UUID_PATH.format(pool=pool, uuid=_uuid), message, ex=self.EXPIRE_TIME)
        self.db.lpush(self.QUEUE_PATH, self.ID_FORMAT.format(pool=pool, uuid=_uuid))

    def pop_next(self, timeout=0):
        res = self.db.brpop(self.QUEUE_PATH, timeout=timeout)
        if res:
            queue_name, log = res
            pool, _uuid = log.decode().split(':', 1)
            return {'pool': pool, 'uuid': _uuid}

    def pop(self, pool, _uuid):
        log_path = self.UUID_PATH.format(pool=pool, uuid=_uuid)
        result = self.db.get(log_path)
        if result:
            self.db.delete(log_path)
        return result




# -*- coding: utf-8 -*-
import os
import signal
import time

import sys

from shire.logger import RedisLogger
from shire.redis_managers import PoolStatusManager, QueueManager
from shire.utils import check_pid_is_shire, create_console_handler, create_logger
from shire.workhorse import Daemon, Workhorse


__all__ = ['Pool']


class Pool(Daemon):
    SLEEP_TIME = 1
    MAX_WORKHORSES = None  # без ограничения

    def __init__(self, config, name, sleep_time=None, max_workhorses=None, verbose=False):
        super(Pool, self).__init__()
        self.config = config
        self.verbose = verbose
        self.redis = config.get_redis()
        self.name = name
        self.pool_logger = create_logger('shire.pool.{}'.format(name))
        if verbose:
            create_console_handler(self.pool_logger)
        self.sleep_time = sleep_time if sleep_time else self.SLEEP_TIME
        self.section = self.config.section_getter(self.config.POOL_SECTION)
        self.check_time = int(self.section.get(self.config.POOL_CHECK_TIME, self.config.POOL_CHECK_TIME_DEFAULT))
        self.max_workhorses = max_workhorses if max_workhorses else self.MAX_WORKHORSES
        self._queue_manager = QueueManager(connection=self.redis)
        self._children = []
        self.log = RedisLogger(config=self.config, pool=self.name)
        self._ps = PoolStatusManager(connection=self.redis)
        # когда форкуется workhorse, то она наследует signal.signal(signal.SIGTERM, self.terminate)
        # и надо проверять в  self.terminate кто сейчас завершается
        self.i_am_pool = True

    def pool_log(self, msg, level='info'):
        if not self.verbose:
            return
        getattr(self.pool_logger, level)(msg)

    def run(self):
        self.pool_log('Pool "{}" started'.format(self.name))
        self.pool_log('Max workhorses: {}'.format('unlimited' if self.max_workhorses is None else self.max_workhorses))
        self.pool_log('Redis check timeout: {}s'.format(self.check_time))
        self._ps.set_status(pool=self.name, _uuid=self.uuid, status=PoolStatusManager.STATUS_ACTIVE)
        signal.signal(signal.SIGTERM, self.terminate)
        while True:
            if not self.can_start_new_workhorse():
                time.sleep(self.sleep_time)
                continue
            while True:
                job_id = self._queue_manager.pop(pool=self.name, timeout=self.check_time)
                if self.status in (PoolStatusManager.STATUS_DEAD, PoolStatusManager.STATUS_KILL):
                    if job_id:
                        self._queue_manager.push(pool=self.name, job_id=job_id, to_tail=True)
                    if self.status == PoolStatusManager.STATUS_KILL:
                        self.kill_children()
                    self.terminate()
                if job_id:
                    # бывает None при выходе из brpop по таймауту
                    self._start_workhorse(job_id)  # тут пролетает sys.exit() от workhorse
                if not self.can_start_new_workhorse():
                    # проверяем после запуска задачи, что бы понять - можно ли выбирать другие задачи
                    break

    def _start_workhorse(self, job_id):
        self.pool_log('Workhorse for job #{} started'.format(job_id))
        workhorse = Workhorse(pool=self, job_id=job_id)
        pid = workhorse.fork()
        self._children.append(pid)

    @property
    def status(self):
        return self._ps.get_status(pool=self.name, _uuid=self.uuid)

    def wait_children(self):
        while self._get_children_count() > 0:
            time.sleep(self.sleep_time)
            if self.status == PoolStatusManager.STATUS_KILL:
                self.kill_children()
                return

    def kill_children(self):
        for child_pid in self._children:
            if check_pid_is_shire(child_pid):
                os.kill(child_pid, signal.SIGKILL)
        self._children = []

    def can_start_new_workhorse(self):
        children_count = self._get_children_count()
        if self.max_workhorses:
            return children_count < self.max_workhorses
        return True

    def _get_children_count(self):
        ended_pids = []
        try:
            while True:
                pid, ret_code = os.waitpid(-1, os.WNOHANG)
                if not pid:
                    break
                ended_pids.append(pid)
        except OSError:
            # нет child-процессов
            pass
        for child_pid in self._children[:]:
            if child_pid in ended_pids:
                self._children.remove(child_pid)
        return len(self._children)

    def terminate(self, sign=None, frame=None):
        if self.i_am_pool:
            self.wait_children()
            self._ps.set_status(pool=self.name, _uuid=self.uuid, status=PoolStatusManager.STATUS_TERMITATED)
        sys.exit(0)



# -*- coding: utf-8 -*-
import datetime
import sys
import time

from shire.exceptions import RestartJobException
from shire.models import JobEntry, Crontab
from shire.pool import Pool
from shire.workhorse import Workhorse

__all__ = ['Job', 'RestartJobException']


class DummyWorkhorse(Workhorse):
    def __init__(self, config):
        pool = Pool(config, 'dummy_pool')
        super(DummyWorkhorse, self).__init__(pool=pool, job_id=None)


def start_job(config, **kwargs):
    with config.with_db():
        job_entry = JobEntry.create_job(**kwargs)
        job_entry.save()
        return job_entry


class Job(object):

    @classmethod
    def delay(cls, config, pool, queue=None, host=None, args=(), kwargs=None,
              sys_path=None, venv_path=None, venv_exclusive=None, wait_minutes=0):
        kwargs = {} if kwargs is None else kwargs
        queue = pool if queue is None else queue
        job_entry = start_job(
            config, pool=pool, queue=queue, host=host,
            file_path=sys.modules[cls.__module__].__file__, file_cls=cls.__name__,
            args=args, kwargs=kwargs,
            sys_path=sys_path, venv_path=venv_path, venv_exclusive=venv_exclusive, wait_minutes=wait_minutes,
        )
        return job_entry

    @classmethod
    def execute(cls, config, args=(), kwargs=None, job_entry=None):
        # Немедленный вызов задачи на выполнение, минуя whip
        kwargs = {} if kwargs is None else kwargs
        # TODO здесь нужны sys_path и venv_path?
        job_entry = start_job(
            config, pool='dummy_pool', queue='dummy_queue', file_path=sys.modules[cls.__module__].__file__,
            file_cls=cls.__name__, args=args, kwargs=kwargs, status=JobEntry.STATUS_ENQUEUED, host='dummy_host'
        ) if job_entry is None else job_entry
        # Т.к. execute однопоточный, то повторять выполнение мы должны прямо здесь.
        # Чтобы не достигнуть maximum recursion depth exceeded раскрываем в while
        while True:
            try:
                job_entry.job_cls(DummyWorkhorse(config)).run(*args, **kwargs)
            except RestartJobException as exc:
                if exc.wait_minutes:
                    # задачка хочет рестартовать отложенно
                    time.sleep(exc.wait_minutes * 60)
                continue
            else:
                break
        return job_entry

    def __init__(self, workhorse):
        self.workhorse = workhorse

    @property
    def log(self):
        return self.workhorse.log

    @property
    def config(self):
        return self.workhorse.config

    @property
    def job_entry(self):
        return self.workhorse.job_entry

    def tick(self):
        # Обновляет информацию о последнем обновлении задачи в базу данных
        self.job_entry.updated_at = datetime.datetime.now()
        self.job_entry.save(only=[JobEntry.updated_at])

    def run(self, *args, **kwargs):
        raise NotImplementedError()

    def restart(self, wait_minutes=None):
        raise RestartJobException(wait_minutes=wait_minutes)

    @classmethod
    def schedule(cls, config, key, cron_string, pool, queue=None, host=None, args=(), kwargs=None,
                 sys_path=None, venv_path=None):
        kwargs = {} if kwargs is None else kwargs
        queue = pool if queue is None else queue
        host = JobEntry.HOST_DEFAULT if host is None else host
        with config.with_db():
            func_call = JobEntry.make_func_call(
                file_path=sys.modules[cls.__module__].__file__,
                file_cls=cls.__name__,
                args=list(args), kwargs=kwargs, sys_path=sys_path, venv_path=venv_path
            )
            crontab = Crontab.select().where(Crontab.key == key).first()
            edited = False
            if crontab is None:
                crontab = Crontab(key=key, cron_string=cron_string, func_call=func_call)
                edited = True
            for key, value in dict(cron_string=cron_string, pool=pool, queue=queue,
                                   host=host, func_call=func_call).items():
                if getattr(crontab, key) != value:
                    setattr(crontab, key, value)
                    edited = True
            if edited:
                # Изменились параметры cron_string, значит надо заново настроить запуски
                crontab.last_scheduled = None
            crontab.save()



# -*- coding: utf-8 -*-
import codecs

import click
import os
import sys
import tabulate

from shire.config import Config
from shire.hostler import Hostler
from shire.models import JobEntry, Queue, Limit, Crontab
from shire.manager import ShireManager
from shire.pool import Pool
from shire.pool_starter import PoolStarter
from shire.scribe import Scribe
from shire.whip import Whip
from shire.utils import to_list, is_venv

__all__ = ['cli']

SHIRE_MODELS = (JobEntry, Queue, Limit, Crontab)  # TODO автогенерить?


@click.group()
@click.option('-c', '--config', type=click.Path(), default='shire.cfg')
@click.option('-v', '--verbose', is_flag=True)
@click.pass_context
def cli(ctx, config, verbose):
    config = os.path.realpath(config)
    cfg = Config()
    ctx.obj = ctx.obj or {}
    if os.path.exists(config):
        cfg.load(config)
        ctx.obj['config_path'] = config
    else:
        cfg.make_default()
        ctx.obj['config_path'] = None
        ctx.obj['requested_config_path'] = config
    ctx.obj.update({'cfg': cfg, 'verbose': verbose})


@cli.command()
@click.pass_context
def init_db(ctx):
    with ctx.obj['cfg'].with_db():
        for model in SHIRE_MODELS:
            model.create_table(fail_silently=True)
            print('Table {} created'.format(model.__name__))


@cli.command()
@click.pass_context
def drop_db(ctx):
    with ctx.obj['cfg'].with_db():
        for model in SHIRE_MODELS:
            model.drop_table(fail_silently=True)
            print('Table {} dropped'.format(model.__name__))


@cli.command()
@click.option('--type', 'db_type_arg', default=None)
@click.option('--name', 'db_name_arg', default=None)
@click.option('--user', 'db_user_arg', default=None)
@click.option('--password', 'db_password_arg', default=None)
@click.option('--host', 'db_host_arg', default=None)
@click.option('--port', 'db_port_arg', default=None)
@click.option('--redis-host', 'redis_host_arg', default=None)
@click.option('--redis-port', 'redis_port_arg', default=None)
@click.option('--redis-db', 'redis_db_arg', default=None)
@click.option('--sys_path', 'sys_path_arg', default=None)
@click.option('--venv_path', 'venv_path_arg', default=None)
@click.option('--force', default=False, is_flag=True)
@click.pass_context
def create_config(
        ctx, db_type_arg, db_name_arg, db_user_arg, db_password_arg, db_host_arg, db_port_arg,
        redis_host_arg, redis_port_arg, redis_db_arg, sys_path_arg, venv_path_arg, force
):
    config_path = ctx.obj['config_path'] or ctx.obj['requested_config_path'] or 'shire.cfg'
    if os.path.exists(config_path):
        if not force and not click.confirm('Config already exists. Overwrite?', default=False):
            return
    db_types = {'0': 'postgres', '1': 'sqlite', '2': 'mysql'}

    if db_type_arg is None:
        result = choices_prompt('Database type', db_types)
        db_type = db_types[result]
    else:
        db_type = db_type_arg
    click.echo('{} choosed.'.format(db_type))
    db_name = (
        click.prompt('Database', default='shire.db' if db_type == 'sqlite' else 'shire_db')
        if db_name_arg is None else db_name_arg
    )
    if db_type == 'sqlite':
        db_url = 'sqlite:///{}'.format(db_name)
    else:
        user = click.prompt('User', default='shire_user') if db_user_arg is None else db_user_arg
        password = click.prompt('Password', hide_input=True) if db_password_arg is None else db_password_arg
        host = click.prompt('Host', default='127.0.0.1') if db_host_arg is None else db_host_arg
        port = (
            click.prompt(u'Port', default='5432' if db_type == 'postgres' else '3306')
            if db_port_arg is None else db_port_arg
        )
        db_url = '{}://{}:{}@{}:{}/{}'.format(db_type, user, password, host, port, db_name)

    redis_host = (
        click.prompt('Redis host', default='127.0.0.1') if redis_host_arg is None else redis_host_arg
    )
    redis_port = (
        click.prompt('Redis port', default='6379') if redis_port_arg is None else redis_port_arg
    )
    redis_db = (
        click.prompt('Redis DB', default='0') if redis_db_arg is None else redis_db_arg
    )
    redis_url = 'redis://{}:{}/{}'.format(redis_host, redis_port, redis_db)

    default_path = os.path.abspath('.')
    sys_path = (
        click.prompt(
            u'PYTHON_PATH for project (Use : as delimiter)', default=default_path, type=click.Path(
                exists=True, file_okay=False, writable=False, readable=True
            )
        ) if sys_path_arg is None else sys_path_arg
    )
    default_venv = os.path.dirname(os.path.dirname(sys.executable)) if is_venv() else ''
    venv_path = (
        click.prompt(
            u'Virtualenv path (Leave blank for disable){}'.format(
                u'[Current venv: {}]'.format(default_venv) if default_venv else ''
            ), default='', type=click.Path(
                exists=True, file_okay=False, writable=False, readable=True
            )
        ) if venv_path_arg is None else venv_path_arg
    )

    # Тут записать все в новый конфиг
    cfg = Config()
    cfg.make_default()
    cfg[Config.SHIRE_SECTION][Config.SHIRE_SYS_PATH] = sys_path
    cfg[Config.SHIRE_SECTION][Config.SHIRE_VENV_PATH] = venv_path
    cfg[Config.CONNECTION_SECTION][Config.CONNECTION_DB_URL] = db_url
    cfg[Config.CONNECTION_SECTION][Config.CONNECTION_REDIS_URL] = redis_url

    config = cfg.to_string_io().getvalue()

    if click.confirm(u'Edit config (some parameters nave been set to default)?', default=True):
        edited = click.edit(config)
        if edited is not None:  # None - если редактирование отменено
            config = edited
    with codecs.open(config_path, 'w', 'utf-8') as f:
        f.write(config)
    click.echo(u'Config successfully saved as {}!'.format(config_path))


@cli.command()
@click.option('-n', '--name', type=click.STRING)
@click.pass_context
def run_pool(ctx, name):
    pool = Pool(config=ctx.obj['cfg'], name=name, verbose=ctx.obj['verbose'])
    pool.run()


@cli.command()
@click.option('-n', '--names', type=click.STRING)
@click.pass_context
def start_pools(ctx, names):
    starter = PoolStarter(config=ctx.obj['cfg'], pools=to_list(names), config_path=ctx.obj['config_path'])
    starter.run()


@cli.command()
@click.pass_context
def run_whip(ctx):
    whip = Whip(config=ctx.obj['cfg'], verbose=ctx.obj['verbose'])
    whip.run()


@cli.command()
@click.pass_context
def run_hostler(ctx):
    hostler = Hostler(config=ctx.obj['cfg'], verbose=ctx.obj['verbose'])
    hostler.run()


@cli.command()
@click.pass_context
def run_scribe(ctx):
    scribe = Scribe(config=ctx.obj['cfg'], verbose=ctx.obj['verbose'])
    scribe.run()


@cli.command()
@click.argument('command')
@click.option('-p', '--pools', type=click.STRING, default=None)
@click.option('-s', '--statuses', type=click.STRING, default=None)
@click.pass_context
def manager(ctx, command, pools, statuses):
    shire_manager = ShireManager(ctx.obj['cfg'])
    kwargs = {}
    if pools is not None:
        kwargs['pools'] = to_list(pools)
    if statuses is not None:
        kwargs['from_statuses'] = to_list(statuses.upper())
    if command == 'get_status':
        click.echo(
            tabulate.tabulate(
                [
                    (x['name'], x['uuid'], x['status'])
                    for x in sorted(shire_manager.run_command(command, **kwargs), key=lambda x: x['name'])
                ],
                headers=['Name', 'UUID', 'Status'], tablefmt='psql'
            )
        )
        return
    shire_manager.run_command(command, **kwargs)


@cli.command()
@click.option('-d', '--days_ago', type=click.INT, default=7)
@click.pass_context
def cleanup_old_jobs(ctx, days_ago):
    with ctx.obj['cfg'].with_db():
        from shire.utils import cleanup_old_jobs
        cleanup_old_jobs(days_ago=days_ago)


@cli.command()
@click.argument('job_id', type=click.INT)
@click.pass_context
def execute_job(ctx, job_id):
    with ctx.obj['cfg'].with_db():
        from shire.utils import execute_job
        result = execute_job(job_id, ctx.obj['cfg'])
        # TODO: Убедиться, что это работает
        return 0 if result else 1


@cli.command()
@click.pass_context
@click.argument('action', type=click.STRING)
@click.option('-p', '--pool', type=click.STRING, default='default')
def crontab(ctx, action, pool):
    from shire.scheduler import CrontabJob
    config = ctx.obj['cfg']
    if action == 'start':
        print('Crontab scheduler started!')
    elif action == 'stop':
        CrontabJob.stop_crontab(config=config)
        print('Crontab scheduler stopped!')
    elif action == 'restart':
        CrontabJob.stop_crontab(config=config)
        CrontabJob.start_crontab(config=config, pool=pool)
        print('Crontab scheduler restarted!')
    elif action == 'clear_jobs':
        CrontabJob.stop_crontab(config=config)
        print('Planned crontab jobs cleared')
    elif action == 'list':
        click.echo(
            tabulate.tabulate(
                [
                    (cron.key, cron.cron_string, cron.last_scheduled or '--', cron.description or '')
                    for cron in CrontabJob.list(config=config)
                ],
                headers=['Key', 'Cron String', 'last_scheduled', 'description'], tablefmt='psql'
            )
        )
    else:
        print('Wrong action! May be start/stop/restart/clear_jobs/list')


def choices_prompt(message, choices):
    return click.prompt(
        u'{}\n{}'.format(
            u'\n'.join([u'{} - {}'.format(k, v) for k, v in sorted(choices.items(), key=lambda x: x[0])]),
            message
        ),
        default='0', type=click.Choice(sorted(choices.keys())), show_default=True
    )

if __name__ == '__main__':
    cli(obj={})



# -*- coding: utf-8 -*-

import datetime

import croniter

from shire.job import Job
from shire.models import Crontab, JobEntry
from shire.utils import create_logger


class CrontabJob(Job):
    SCHEDULE_DELTA_MINUTES = 60
    RERUN_PERIOD_MINUTES = 30
    CRONTAB_QUEUE = '__shire_crontab__'

    def __init__(self, workhorse):
        self.logger = create_logger('shire.schedule')
        super(CrontabJob, self).__init__(workhorse=workhorse)

    def run(self):
        start = datetime.datetime.now()
        for crontab in Crontab.select():
            schedule_from = crontab.last_scheduled if crontab.last_scheduled else start
            schedule_until = start + datetime.timedelta(minutes=self.SCHEDULE_DELTA_MINUTES)
            if schedule_until <= schedule_from:
                continue
            croniter_obj = croniter.croniter(
                expr_format=crontab.cron_string,
                start_time=schedule_from,
                ret_type=datetime.datetime
            )
            next_run = croniter_obj.get_next()
            while next_run <= schedule_until:
                try:
                    JobEntry.create(
                        func_call_=crontab.func_call_,
                        pool=crontab.pool,
                        queue=crontab.queue,
                        host=crontab.host,
                        execute_at=next_run,
                    )
                    crontab.last_scheduled = next_run
                    crontab.save()
                except Exception as e:
                    self.logger.error(u'Ошибка во время запуска задачи #%s', crontab.id)
                    self.logger.exception(e)
                else:
                    self.logger.info(u'Задача #%s успешно поставлена на выполнение в %s', crontab.id, next_run)
                next_run = croniter_obj.get_next()

        future_crontab_jobs = list(JobEntry.select().where(
            (JobEntry.execute_at > start) &
            (JobEntry.queue == CrontabJob.CRONTAB_QUEUE)
        ).order_by(JobEntry.execute_at))
        if len(future_crontab_jobs):
            last_job = future_crontab_jobs.pop()
            last_job.execute_at = start + datetime.timedelta(minutes=self.RERUN_PERIOD_MINUTES)
            last_job.save()
            for job in future_crontab_jobs:
                job.delete_instance()
        else:
            self.start_crontab(config=self.workhorse.config, pool=self.workhorse.pool.name,
                               wait_minutes=self.RERUN_PERIOD_MINUTES)

    @classmethod
    def start_crontab(cls, *args, **kwargs):
        kwargs['queue'] = cls.CRONTAB_QUEUE
        return cls.delay(*args, **kwargs)

    @classmethod
    def stop_crontab(cls, config):
        with config.with_db():
            JobEntry.delete().where(JobEntry.queue == cls.CRONTAB_QUEUE).execute()

    @classmethod
    def clear_jobs(cls, config):
        with config.with_db():
            JobEntry.delete().where(
                (JobEntry.execute_at > datetime.datetime.now()) &
                (JobEntry.queue != cls.CRONTAB_QUEUE)
            ).execute()

    @classmethod
    def list(cls, config):
        with config.with_db():
            return Crontab.select()



# -*- coding: utf-8 -*-
import codecs

import six
from six.moves import configparser

from playhouse.db_url import connect
from redis import StrictRedis


__all__ = ['Config']


class Config(dict):
    SHIRE_SECTION = 'shire'
    SHIRE_HOST = 'host'
    SHIRE_HOST_DEFAULT = 'default'
    SHIRE_SYS_PATH = 'sys_path'
    SHIRE_SYS_PATH_DEFAULT = ''
    SHIRE_VENV_PATH = 'venv_path'
    SHIRE_VENV_PATH_DEFAULT = ''
    SHIRE_VENV_EXCLUSIVE = 'venv_path'
    SHIRE_VENV_EXCLUSIVE_DEFAULT = None

    CONNECTION_SECTION = 'connection'
    CONNECTION_DB_URL = 'db_url'
    CONNECTION_DB_URL_DEFAULT = 'sqlite:///shire.db'
    CONNECTION_REDIS_URL = 'redis_url'
    CONNECTION_REDIS_URL_DEFAULT = 'redis://localhost:6379/0'

    POOL_SECTION = 'pool'
    POOL_CHECK_TIME = 'check_time'
    POOL_CHECK_TIME_DEFAULT = '30'
    
    SCRIBE_SECTION = 'scribe'
    SCRIBE_PER_POOL = 'per_pool'
    SCRIBE_PER_POOL_DEFAULT = '1'
    SCRIBE_DIRECTORY = 'log_directory'
    SCRIBE_DIRECTORY_DEFAULT = 'logs'
    SCRIBE_SYSTEM_LOG = 'system_log'
    SCRIBE_SYSTEM_LOG_DEFAULT = 'shire.log'
    SCRIBE_DEFAULT_LOG = 'default_log'
    SCRIBE_DEFAULT_LOG_DEFAULT = 'default.log'
    SCRIBE_MAX_SIZE = 'log_max_size'
    SCRIBE_MAX_SIZE_DEFAULT = str(100 * 1024 * 1024)  # 100 мегабайт
    SCRIBE_MAX_LOGS = 'max_logs'
    SCRIBE_MAX_LOGS_DEFAULT = '5'

    HOSTLER_SECTION = 'hostler'
    HOSTLER_CHECK_TIME = 'check_time'
    HOSTLER_CHECK_TIME_DEFAULT = '5'

    WHIP_SECTION = 'whip'
    WHIP_CHECK_TIME = 'check_time'
    WHIP_CHECK_TIME_DEFAULT = '1'
    WHIP_LIMITS_UPDATE_TIME = 'limits_update_time'
    WHIP_LIMITS_UPDATE_TIME_DEFAULT = '60'
    WHIP_MAX_JOBS = 'max_jobs'
    WHIP_MAX_JOBS_DEFAULT = '100'

    def __init__(self, *args, **kwargs):
        self.cp = configparser.ConfigParser()
        super(Config, self).__init__(*args, **kwargs)

    def make_default(self):
        self.clear()
        self.update({
            self.SHIRE_SECTION: {
                self.SHIRE_HOST: self.SHIRE_HOST_DEFAULT,
                self.SHIRE_SYS_PATH: self.SHIRE_SYS_PATH_DEFAULT,
                self.SHIRE_VENV_PATH: self.SHIRE_VENV_PATH_DEFAULT
            },
            self.CONNECTION_SECTION: {
                self.CONNECTION_DB_URL: self.CONNECTION_DB_URL_DEFAULT,
                self.CONNECTION_REDIS_URL: self.CONNECTION_REDIS_URL_DEFAULT
            },
            self.POOL_SECTION: {
                self.POOL_CHECK_TIME: self.POOL_CHECK_TIME_DEFAULT
            },
            self.SCRIBE_SECTION: {
                self.SCRIBE_PER_POOL: self.SCRIBE_PER_POOL_DEFAULT,
                self.SCRIBE_DIRECTORY: self.SCRIBE_DIRECTORY_DEFAULT,
                self.SCRIBE_DEFAULT_LOG: self.SCRIBE_DEFAULT_LOG_DEFAULT,
                self.SCRIBE_MAX_SIZE: self.SCRIBE_MAX_SIZE_DEFAULT,
                self.SCRIBE_MAX_LOGS: self.SCRIBE_MAX_LOGS_DEFAULT
            },
            self.HOSTLER_SECTION: {
                self.HOSTLER_CHECK_TIME: self.HOSTLER_CHECK_TIME_DEFAULT
            },
            self.WHIP_SECTION: {
                self.WHIP_CHECK_TIME: self.WHIP_CHECK_TIME_DEFAULT,
                self.WHIP_LIMITS_UPDATE_TIME: self.WHIP_LIMITS_UPDATE_TIME_DEFAULT
            }
        })
        self.save_to_cp()

    def load(self, file_path):
        self.cp = configparser.ConfigParser()
        self.cp.read(file_path)
        self.load_from_cp()

    def load_from_cp(self):
        for section in self.cp.sections():
            for option in self.cp.options(section):
                value = self.cp.get(section, option)
                self.setdefault(section, {})[option] = value

    def save(self, file_path):
        self.save_to_cp()
        self.cp.write(codecs.open(file_path, 'w', 'utf-8'))

    def save_to_cp(self):
        self.cp = configparser.ConfigParser()
        for section, variables in self.items():
            self.cp.add_section(section)
            for name, value in variables.items():
                self.cp.set(section, name, value)

    def get_db(self):
        url = self.get(self.CONNECTION_SECTION, {}).get(self.CONNECTION_DB_URL, None)
        if url is None:
            return None
        return connect(url)

    class _with_db(object):
        def __init__(self, database_getter):
            from shire.models import db
            try:
                is_closed = db.is_closed()
            except AttributeError:
                # Это неинициализированный Proxy
                is_closed = True
            self.need_manage = is_closed
            self.db = db
            if is_closed:
                db.initialize(database_getter())

        def __enter__(self):
            if self.need_manage:
                self.db.connect()

        def __exit__(self, exc_type, exc_val, exc_tb):
            if self.need_manage:
                self.db.close()

    def with_db(self):
        return self._with_db(self.get_db)

    def get_redis(self):
        url = self.get(self.CONNECTION_SECTION, {}).get(self.CONNECTION_REDIS_URL, None)
        if url is None:
            return None
        return StrictRedis.from_url(url)

    def from_section(self, section, key, default=None):
        return self.get(section, {}).get(key, default)

    def section_getter(self, section):
        return type('_', (object,), {'get': lambda s, key, default=None: self.from_section(section, key, default)})()

    def to_string_io(self):
        string = six.StringIO()
        self.save_to_cp()
        self.cp.write(string)
        return string



# -*- coding: utf-8 -*-

import collections
import time

import datetime

from shire.models import db, JobEntry, Limit, Queue
from shire.redis_managers import QueueManager
from shire.utils import create_console_handler, create_logger


__all__ = ['Whip']


class Whip(object):
    # Специально написан по методологии "let it crash".
    # В случае непредвиденных ситуаций должен быть перезапущен супервайзером, а не пытаться разрешить их самостоятельно,
    # в ущерб стабильности

    def __init__(self, config, verbose=False):
        self.config = config
        self.verbose = verbose
        self.whip_logger = create_logger('shire.whip')
        if verbose:
            create_console_handler(self.whip_logger)
        self.section = self.config.section_getter(self.config.WHIP_SECTION)
        db.initialize(self.config.get_db())
        self.queue_limits = {}
        self.pool_limits = {}
        self.queue_to_pool = {}
        self.max_jobs_limit = int(self.section.get(self.config.WHIP_MAX_JOBS, self.config.WHIP_MAX_JOBS_DEFAULT))
        self.redis_queue = QueueManager(self.config.get_redis())
        self.host = self.section.get(self.config.SHIRE_HOST, self.config.SHIRE_HOST_DEFAULT)

    def whip_log(self, msg, level='info'):
        if not self.verbose:
            return
        getattr(self.whip_logger, level)(msg)

    def update_limits(self):
        self.queue_to_pool = {x.name: x.pool for x in Queue.select()}
        for limit in Limit.select():
            if limit.entity_type == Limit.ENTITY_QUEUE:
                self.queue_limits[limit.entity] = limit.limit
            elif limit.entity_type == Limit.ENTITY_POOL:
                self.pool_limits[limit.entity] = limit.limit
        self.whip_log('Limits updated')

    def load_current_jobs(self):
        total = 0
        by_pool = collections.defaultdict(int)
        by_queue = collections.defaultdict(int)

        for job_entry in JobEntry.select(
                JobEntry.id, JobEntry.pool, JobEntry.queue
        ).where(JobEntry.status << [JobEntry.STATUS_IN_PROGRESS, JobEntry.STATUS_ENQUEUED]):
            total += 1
            by_pool[job_entry.pool] += 1
            by_queue[job_entry.queue] += 1

        return {
            'total': total,
            'by_pool': by_pool,
            'by_queue': by_queue
        }

    def enqueue_job(self, job_entry):
        self.whip_log('Job #{} enqueued (Pool: {})'.format(job_entry.id, job_entry.pool))
        self.redis_queue.push(pool=job_entry.pool, job_id=job_entry.id)
        job_entry.status = JobEntry.STATUS_ENQUEUED
        job_entry.save(only=[JobEntry.status])

    def can_enqueue(self, job_entry, current):
        if current['total'] >= self.max_jobs_limit:
            return False
        limit_by_queue = current['by_queue'].get(job_entry.queue, 0)
        if job_entry.queue in self.queue_limits and limit_by_queue >= self.queue_limits[job_entry.queue]:
            return False
        limit_by_pool = current['by_pool'].get(job_entry.pool, 0)
        if job_entry.pool in self.pool_limits and limit_by_pool >= self.pool_limits[job_entry.pool]:
            return False
        return True

    def run(self):
        last_update_limits = 0
        update_limits_time = int(self.section.get(
            self.config.WHIP_LIMITS_UPDATE_TIME, self.config.WHIP_LIMITS_UPDATE_TIME_DEFAULT
        ))
        time_to_sleep = int(self.section.get(self.config.WHIP_CHECK_TIME, self.config.WHIP_CHECK_TIME_DEFAULT))
        loop_sleep_time = 0
        self.whip_log('Whip stared')
        self.whip_log('Jobs check time: {}s'.format(time_to_sleep))
        self.whip_log('Update queue limits time: {}s'.format(update_limits_time))

        while True:
            next_run = time.time() + time_to_sleep
            if loop_sleep_time > 0:
                # Спим только если предыдущее выполнение было короче time_to_sleep
                time.sleep(loop_sleep_time)

            # Обновляем лимиты, если пришо время
            if last_update_limits < (time.time() - update_limits_time):
                self.update_limits()
                last_update_limits = time.time()

            # Получаем информацию о текущих задачах
            current = self.load_current_jobs()

            if current['total'] >= self.max_jobs_limit:
                # Превышено максимальное количество задач
                continue

            # Отправляем на выполнение новые задачи
            for job_entry in JobEntry.select(
                JobEntry.id, JobEntry.pool, JobEntry.queue, JobEntry.status
            ).where(
                (JobEntry.status << [JobEntry.STATUS_NEW, JobEntry.STATUS_RESTART]) 
                & (JobEntry.host == self.host)
                & (JobEntry.execute_at <= datetime.datetime.now())
            ).order_by(JobEntry.execute_at.asc()):
                if self.can_enqueue(job_entry=job_entry, current=current):
                    self.enqueue_job(job_entry=job_entry)
                    current['total'] += 1
                    current['by_queue'][job_entry.queue] += 1
                    current['by_pool'][job_entry.pool] += 1

            # Компенсируем время, затраченное на выполнение
            loop_sleep_time = next_run - time.time()



# -*- coding: utf-8 -*-


__all__ = ['ShireException', 'PoolInvalidStatusException', 'RestartJobException']


class ShireException(Exception):
    pass


class PoolInvalidStatusException(ShireException):
    pass


class RestartJobException(Exception):

    def __init__(self, wait_minutes=None, *args):
        self.wait_minutes = wait_minutes
        super(RestartJobException, self).__init__(*args)




# -*- coding: utf-8 -*-
import setproctitle

import six
import os
import uuid
import sys

import datetime
import warnings

from shire.const import SHIRE_WORKHORSE_PROCESS_NAME
from shire.exceptions import RestartJobException
from shire.models import JobEntry, db
from shire.utils import capture_output, activate_venv


__all__ = ['Workhorse', 'WorkhorseStatusContext']


class WorkhorseStatusContext(object):

    def __init__(self, workhorse):
        self.workhorse = workhorse

    def __enter__(self):
        job_entry = self.workhorse.job_entry
        job_entry.pool_uuid = self.workhorse.pool.uuid
        job_entry.worker_uuid = self.workhorse.uuid
        job_entry.worker_pid = self.workhorse.pid
        job_entry.status = JobEntry.STATUS_IN_PROGRESS
        job_entry.save(only=[JobEntry.status, JobEntry.worker_uuid, JobEntry.worker_pid, JobEntry.pool_uuid])
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        exceptions_proceeded = False
        job_entry = self.workhorse.job_entry
        if exc_type == RestartJobException:
            if exc_val.wait_minutes:
                # задачка хочет рестартовать отложенно
                job_entry.execute_at = datetime.datetime.now() + datetime.timedelta(minutes=exc_val.wait_minutes)
            job_entry.status = JobEntry.STATUS_RESTART
            exceptions_proceeded = True
        else:
            job_entry.status = JobEntry.STATUS_ENDED
        job_entry.save(only=[JobEntry.status, JobEntry.execute_at])
        return exceptions_proceeded


class Daemon(object):

    def __init__(self):
        self._uuid = None
        self._pid = None

    @property
    def uuid(self):
        if not self._uuid:
            self._uuid = str(uuid.uuid4())
        return self._uuid

    @property
    def pid(self):
        if not self._pid:
            self._pid = os.getpid()
        return self._pid


class Workhorse(Daemon):

    def __init__(self, pool, job_id):
        super(Workhorse, self).__init__()
        self.pool = pool
        self.config = pool.config
        self.log = pool.log
        self.job_id = job_id
        self.job_entry = None

    @property
    def sys_path(self):
        sys_path = self.config.get(self.config.SHIRE_SECTION, {}).get(
            self.config.SHIRE_SYS_PATH, self.config.SHIRE_SYS_PATH_DEFAULT
        )
        if not sys_path:
            return None

        return [os.path.abspath(x) for x in sys_path.split(':')]

    @property
    def venv_path(self):
        venv_path = self.config.get(self.config.SHIRE_SECTION, {}).get(
            self.config.SHIRE_VENV_PATH, self.config.SHIRE_VENV_PATH_DEFAULT
        )
        if not venv_path:
            return None
        return os.path.abspath(venv_path)

    @property
    def venv_exclusive(self):
        venv_exclusive = self.config.get(self.config.SHIRE_SECTION, {}).get(
            self.config.SHIRE_VENV_EXCLUSIVE, self.config.SHIRE_VENV_EXCLUSIVE_DEFAULT
        )
        return venv_exclusive

    def fork(self):
        pid = os.fork()
        if pid:
            return pid
        # унаследовали signal.signal(signal.SIGTERM, pool.terminate)
        # надо сказать инстансу пула, что мы теперь - workhorse
        self.pool.i_am_pool = False
        setproctitle.setproctitle('{} job_id={}'.format(SHIRE_WORKHORSE_PROCESS_NAME, self.job_id) + ' ' * 128)
        # отдельное соединение для потомка
        db.initialize(self.config.get_db())
        try:
            self.run()
        except Exception as exc:
            self.log.exception('In job #{} with workhorse uuid {} (pool {}/{})'.format(
                self.job_id, self.uuid, self.pool.name, self.pool.uuid))
            sys.exit(1)
        else:
            sys.exit(0)
        finally:
            db.close()

    def run(self):
        self.job_entry = self._load_from_db()
        # Активация virtual_env
        job_venv_path = self.job_entry.func_call.get(JobEntry.FUNC_CALL_VENV_PATH, None)
        if job_venv_path is not None:
            job_venv_exclusive = self.job_entry.func_call.get(JobEntry.FUNC_CALL_VENV_EXCLUSIVE, None)
            activate_venv(job_venv_path, job_venv_exclusive)
        elif self.venv_path:
            activate_venv(self.venv_path, self.venv_exclusive)
        # Обновление sys_path
        if self.job_entry.func_call.get(JobEntry.FUNC_CALL_SYS_PATH) is not None:
            sys_path = self.job_entry.func_call[JobEntry.FUNC_CALL_SYS_PATH]
            sys_path = [os.path.abspath(x) for x in sys_path.split(':')]
            for path in reversed(sys_path):
                if path not in sys.path:
                    sys.path.insert(0, path)
        elif self.sys_path:
            for path in reversed(self.sys_path):
                if path not in sys.path:
                    sys.path.insert(0, path)

        stream = six.StringIO()
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            with capture_output(stream, stream):
                job = self.job_entry.job_cls(workhorse=self)
                args, kwargs = self.job_entry.get_params()
                with WorkhorseStatusContext(workhorse=self):
                    job.run(*args, **kwargs)
        output = stream.getvalue() or None
        if output is not None:
            try:
                self.log.info(u'Got output for job #{}: \n{}'.format(self.job_id, output))
            except UnicodeDecodeError:
                pass

    def _load_from_db(self):
        job_entry = JobEntry.get(JobEntry.id == self.job_id)
        return job_entry



# -*- coding: utf-8 -*-
import os
import signal
import sys
import time

import setproctitle

from shire.manager import ShireManager
from shire.pool import Pool


class PoolStarter(object):

    def __init__(self, config, pools=(), config_path=''):
        self.config = config
        self.pools = pools
        self.config_path = config_path

    def run(self):
        signal.signal(signal.SIGTERM, self.terminate)
        signal.signal(signal.SIGINT, self.terminate)
        for pool_name in self.pools:
            self.spawn_child(pool_name=pool_name)
        try:
            while True:
                time.sleep(1)
        except Exception as exc:
            self.terminate()

    def spawn_child(self, pool_name):
        # see https://stackoverflow.com/questions/972362/spawning-process-from-python/972383#972383

        try:
            pid = os.fork()
        except OSError as e:
            raise RuntimeError("1st fork failed: %s [%d]" % (e.strerror, e.errno))
        if pid != 0:
            # parent (calling) process is all done
            return pid
        # procname.setprocname('python -m shire.cli pool_starter FIRST child' + ' ' * 512)

        # detach from controlling terminal (to make child a session-leader)
        os.setsid()
        try:
            # Fork a second child and exit immediately to prevent zombies.  This
            # causes the second child process to be orphaned, making the init
            # process responsible for its cleanup.  And, since the first child is
            # a session leader without a controlling terminal, it's possible for
            # it to acquire one by opening a terminal in the future (System V-
            # based systems).  This second fork guarantees that the child is no
            # longer a session leader, preventing the daemon from ever acquiring
            # a controlling terminal.
            pid = os.fork()
        except OSError as e:
            raise RuntimeError("2nd fork failed: %s [%d]" % (e.strerror, e.errno))
        if pid != 0:
            # child process is all done
            os._exit(0)
        # procname.setprocname('python -m shire.cli pool_starter SECOND child' + ' ' * 512)

        # grandchild process now non-session-leader, detached from parent
        # grandchild process must now close all open files
        try:
            maxfd = os.sysconf("SC_OPEN_MAX")
        except (AttributeError, ValueError):
            maxfd = 1024

        for fd in range(maxfd):
            try:
                os.close(fd)
            except OSError:  # ERROR, fd wasn't open to begin with (ignored)
                pass

        # redirect stdin, stdout and stderr to /dev/null
        os.open(os.devnull, os.O_RDWR)  # standard input (0)
        os.dup2(0, 1)
        os.dup2(0, 2)

        # and finally let's execute the executable for the daemon!
        cmd = 'shire.cli --config={} run_pool --name {}'.format(self.config_path, pool_name)
        setproctitle.setproctitle(cmd + ' ' * 512)
        pool = Pool(config=self.config, name=pool_name)
        try:
            pool.run()
        except Exception as e:
            pool.log.exception('Pool name {} uuid{}'.format(pool.name, pool.uuid))
            os._exit(1)
        else:
            os._exit(0)

    def terminate(self, sign=None, frame=None):
        manager = ShireManager(config=self.config)
        manager.terminate_pools(self.pools)
        sys.exit(0)




# -*- coding: utf-8 -*-

import logging
import sys
import traceback
import json

import six

from shire.redis_managers import LogMessageManager


__all__ = ['RedisLogger']


class RedisLogger(logging.Logger):
    # Специальный логгер shire, не требующий подключения handler'ов

    def __init__(self, config, pool='shire'):
        self.shire_config = config
        self.shire_log_manager = LogMessageManager(config.get_redis())
        super(RedisLogger, self).__init__(pool, level=logging.NOTSET)

    @classmethod
    def format_exception(cls, ei):
        sio = six.StringIO()
        traceback.print_exception(ei[0], ei[1], ei[2], None, sio)
        s = sio.getvalue()
        sio.close()
        if s[-1:] == "\n":
            s = s[:-1]
        return s

    @classmethod
    def format_record(cls, record):
        # Переработанный обработчик из logging.Formatter.format
        record.message = record.getMessage()
        if record.exc_info:
            if not record.exc_text:
                record.exc_text = cls.format_exception(record.exc_info)
        if record.exc_text:
            try:
                '' + record.exc_text
            except UnicodeError:
                record.exc_text = record.exc_text.decode(sys.getfilesystemencoding(), 'replace')
        del record.exc_info
        del record.msg
        return json.dumps(record.__dict__)

    def handle(self, record):
        # Вместо вызова хенделоров сразу пишем в редис
        json_str = self.format_record(record)
        self.shire_log_manager.write(pool=self.name, message=json_str)



# -*- coding: utf-8 -*-
import datetime
import time
import sys

from shire.models import db, JobEntry
from shire.utils import check_pid_is_shire, create_console_handler, create_logger


__all__ = ['Hostler']


class Hostler(object):
    # Специально написан по методологии "let it crash".
    # В случае непредвиденных ситуаций должен быть перезапущен супервайзером, а не пытаться разрешить их самостоятельно,
    # в ущерб стабильности

    CHECK_MINUTES = 5  # Задачи не обновлявшиеся с какого времени проверять

    def __init__(self, config, verbose=False):
        self.config = config
        self.verbose = verbose
        self.hostler_logger = create_logger('shire.hostler')
        if verbose:
            create_console_handler(self.hostler_logger)
        self.section = self.config.section_getter(self.config.HOSTLER_SECTION)
        db.initialize(self.config.get_db())
        self.host = self.config.get(self.config.SHIRE_SECTION, {}).get(
            self.config.SHIRE_HOST, self.config.SHIRE_HOST_DEFAULT
        )
        self.already_checked = {}

    def hostler_log(self, msg, level='info'):
        if not self.verbose:
            return
        getattr(self.hostler_logger, level)(msg)

    def restart_job(self, job_entry):
        self.hostler_log('Task #{} restarted'.format(job_entry.id))
        job_entry.status = JobEntry.STATUS_RESTART
        job_entry.save(only=[JobEntry.status])

    def loop(self):
        last_updated = datetime.datetime.now() - datetime.timedelta(minutes=self.CHECK_MINUTES)
        for job_entry in JobEntry.select(JobEntry.id, JobEntry.status, JobEntry.worker_pid).where(
                                (JobEntry.status == JobEntry.STATUS_IN_PROGRESS) &
                                (JobEntry.updated_at < last_updated) & (JobEntry.host == self.host)
        ):
            if job_entry.id not in self.already_checked or self.already_checked[job_entry.id] < last_updated:
                if not check_pid_is_shire(job_entry.worker_pid):
                    self.restart_job(job_entry)
                self.already_checked[job_entry.id] = datetime.datetime.now()
        time.sleep(float(
            self.section.get(self.config.HOSTLER_CHECK_TIME, self.config.HOSTLER_CHECK_TIME_DEFAULT)
        ))

    def run(self):
        self.hostler_log('Hostler started')
        while True:
            try:
                self.loop()
            except Exception as e:
                self.hostler_log(e, 'exception')
                sys.exit(1)



# -*- coding: utf-8 -*-

from shire.redis_managers import PoolStatusManager


__all__ = ['ShireManager']


class ShireManager(object):
    CONST_ALL = '__all__'
    EXPORT_METHODS = (
        'get_status', 'terminate_pools', 'kill_pools', 'clean_pools'
    )

    def __init__(self, config):
        self.config = config
        self.status_manager = PoolStatusManager(self.config.get_redis())

    def _prepare_statuses(self, statuses):
        if statuses == self.CONST_ALL:
            statuses = (
                PoolStatusManager.STATUS_DEAD, PoolStatusManager.STATUS_KILL,
                PoolStatusManager.STATUS_ACTIVE, PoolStatusManager.STATUS_TERMITATED
            )
        return set(statuses)

    def get_pools(self):
        # TODO: Добавить опциональный параметр get_extra=False, при получении которого загружать так же из
        # базы данных кол-во активных, ожидающих и выполненных задач, а так же время последнего обновления задачи
        for key in self.status_manager.get_all():
            try:
                prefix, pool, uuid = key.rsplit(':', 2)
            except ValueError:
                # Плохое значение в get_all. Кейс скорее невозможный, проверка на всякий
                continue
            yield {
                'name': pool,
                'uuid': uuid,
                'status': self.status_manager.get_status(pool, uuid)
            }

    def get_status(self, pools=CONST_ALL, from_statuses=CONST_ALL):
        statuses = self._prepare_statuses(from_statuses)
        check_name = (lambda x: True) if pools == self.CONST_ALL else (lambda x: x in pools)
        for pool in self.get_pools():
            if pool['status'] in statuses and check_name(pool['name']):
                yield pool

    def update_pool_statuses(self, status, pools=CONST_ALL, from_statuses=CONST_ALL):
        statuses = self._prepare_statuses(from_statuses)
        check_name = (lambda x: True) if pools == self.CONST_ALL else (lambda x: x in pools)
        for pool in self.get_pools():
            if pool['status'] in statuses and check_name(pool['name']):
                self.status_manager.set_status(pool['name'], pool['uuid'], status)

    def terminate_pools(self, pools=CONST_ALL, from_statuses=(PoolStatusManager.STATUS_ACTIVE,)):
        # Останавливаем пулы, посылая им статус DEAD
        self.update_pool_statuses(
            PoolStatusManager.STATUS_DEAD, pools=pools, from_statuses=from_statuses
        )

    def kill_pools(self, pools=CONST_ALL, from_statuses=(PoolStatusManager.STATUS_DEAD,)):
        # Приказывает пулам принудительно уничтожить всех своих потомков, посылая им статус KILL
        self.update_pool_statuses(
            PoolStatusManager.STATUS_KILL, pools=pools, from_statuses=from_statuses
        )

    def clean_pools(self, from_statuses=(PoolStatusManager.STATUS_TERMITATED,), **kwargs):
        # Очищает записи в redis от устаревших uuid
        statuses = self._prepare_statuses(from_statuses)
        for pool in self.get_pools():
            if pool['status'] in statuses:
                self.status_manager.del_status(pool['name'], pool['uuid'])

    def run_command(self, command, *args, **kwargs):
        if command in self.EXPORT_METHODS:
            return getattr(self, command)(*args, **kwargs)



# -*- coding: utf-8 -*-

import codecs
import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

with codecs.open('requirements.txt', 'r', 'utf-8') as f:
    requirements = [x.strip() for x in f.read().splitlines() if x.strip()]

setup(
    name='shire',
    version='0.2.0',
    description='Heavy-job queue processor',
    long_description=long_description,
    author='Suguby',
    author_email='suguby@gmail.com',
    url='https://bitbucket.org/suguby/shire',
    license='MIT',
    packages=find_packages(include=['shire']),
    data_files=[('.', ['requirements.txt', 'README.rst'], ), ],
    install_requires=requirements,
    entry_points={
        'console_scripts': ['shire-cli=shire.cli:cli']
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Natural Language :: Russian',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='queue job task shire',
)
